Imports System.Collections
Imports System.Diagnostics.CodeAnalysis
Imports System.Globalization
Imports System.Reflection

Namespace GovDataAPI.Areas.HelpPage
	''' <summary>
	''' This class will create an object of a given type and populate it with sample data.
	''' </summary>
	Public Class ObjectGenerator
		Private Const DefaultCollectionSize As Integer = 3
		Private ReadOnly SimpleObjectGenerator As New SimpleTypeObjectGenerator()

		''' <summary>
		''' Generates an object for a given type. The type needs to be public, have a public default constructor and settable public properties/fields. Currently it supports the following types:
		''' Simple types: <see cref="int"/>, <see cref="string"/>, <see cref="Enum"/>, <see cref="DateTime"/>, <see cref="Uri"/>, etc.
		''' Complex types: POCO types.
		''' Nullables: <see cref="Nullable{T}"/>.
		''' Arrays: arrays of simple types or complex types.
		''' Key value pairs: <see cref="KeyValuePair{TKey,TValue}"/>
		''' Tuples: <see cref="Tuple{T1}"/>, <see cref="Tuple{T1,T2}"/>, etc
		''' Dictionaries: <see cref="IDictionary{TKey,TValue}"/> or anything deriving from <see cref="IDictionary{TKey,TValue}"/>.
		''' Collections: <see cref="IList{T}"/>, <see cref="IEnumerable{T}"/>, <see cref="ICollection{T}"/>, <see cref="IList"/>, <see cref="IEnumerable"/>, <see cref="ICollection"/> or anything deriving from <see cref="ICollection{T}"/> or <see cref="IList"/>.
		''' Queryables: <see cref="IQueryable"/>, <see cref="IQueryable{T}"/>.
		''' </summary>
		''' <param name="type">The type.</param>
		''' <returns>An object of the given type.</returns>
		Public Function GenerateObject(ByVal type As Type) As Object
			Return GenerateObject(type, New Dictionary(Of Type, Object)())
		End Function

		<SuppressMessage("Microsoft.Design", "CA1031:DoNotCatchGeneralExceptionTypes", Justification := "Here we just want to return null if anything goes wrong.")>
		Private Function GenerateObject(ByVal type As Type, ByVal createdObjectReferences As Dictionary(Of Type, Object)) As Object
			Try
				If SimpleTypeObjectGenerator.CanGenerateObject(type) Then
					Return SimpleObjectGenerator.GenerateObject(type)
				End If

				If type.IsArray Then
					Return GenerateArray(type, DefaultCollectionSize, createdObjectReferences)
				End If

				If type.IsGenericType Then
					Return GenerateGenericType(type, DefaultCollectionSize, createdObjectReferences)
				End If

				If type Is GetType(IDictionary) Then
					Return GenerateDictionary(GetType(Hashtable), DefaultCollectionSize, createdObjectReferences)
				End If

				If GetType(IDictionary).IsAssignableFrom(type) Then
					Return GenerateDictionary(type, DefaultCollectionSize, createdObjectReferences)
				End If

				If type Is GetType(IList) OrElse type Is GetType(IEnumerable) OrElse type Is GetType(ICollection) Then
					Return GenerateCollection(GetType(ArrayList), DefaultCollectionSize, createdObjectReferences)
				End If

				If GetType(IList).IsAssignableFrom(type) Then
					Return GenerateCollection(type, DefaultCollectionSize, createdObjectReferences)
				End If

				If type Is GetType(IQueryable) Then
					Return GenerateQueryable(type, DefaultCollectionSize, createdObjectReferences)
				End If

				If type.IsEnum Then
					Return GenerateEnum(type)
				End If

				If type.IsPublic OrElse type.IsNestedPublic Then
					Return GenerateComplexObject(type, createdObjectReferences)
				End If
			Catch
				' Returns null if anything fails
				Return Nothing
			End Try

			Return Nothing
		End Function

		Private Shared Function GenerateGenericType(ByVal type As Type, ByVal collectionSize As Integer, ByVal createdObjectReferences As Dictionary(Of Type, Object)) As Object
			Dim genericTypeDefinition As Type = type.GetGenericTypeDefinition()
			If genericTypeDefinition Is GetType(Nullable(Of )) Then
				Return GenerateNullable(type, createdObjectReferences)
			End If

			If genericTypeDefinition Is GetType(KeyValuePair(Of ,)) Then
				Return GenerateKeyValuePair(type, createdObjectReferences)
			End If

			If IsTuple(genericTypeDefinition) Then
				Return GenerateTuple(type, createdObjectReferences)
			End If

			Dim genericArguments() As Type = type.GetGenericArguments()
			If genericArguments.Length = 1 Then
				If genericTypeDefinition Is GetType(IList(Of )) OrElse genericTypeDefinition Is GetType(IEnumerable(Of )) OrElse genericTypeDefinition Is GetType(ICollection(Of )) Then
					Dim collectionType As Type = GetType(List(Of )).MakeGenericType(genericArguments)
					Return GenerateCollection(collectionType, collectionSize, createdObjectReferences)
				End If

				If genericTypeDefinition Is GetType(IQueryable(Of )) Then
					Return GenerateQueryable(type, collectionSize, createdObjectReferences)
				End If

				Dim closedCollectionType As Type = GetType(ICollection(Of )).MakeGenericType(genericArguments(0))
				If closedCollectionType.IsAssignableFrom(type) Then
					Return GenerateCollection(type, collectionSize, createdObjectReferences)
				End If
			End If

			If genericArguments.Length = 2 Then
				If genericTypeDefinition Is GetType(IDictionary(Of ,)) Then
					Dim dictionaryType As Type = GetType(Dictionary(Of ,)).MakeGenericType(genericArguments)
					Return GenerateDictionary(dictionaryType, collectionSize, createdObjectReferences)
				End If

				Dim closedDictionaryType As Type = GetType(IDictionary(Of ,)).MakeGenericType(genericArguments(0), genericArguments(1))
				If closedDictionaryType.IsAssignableFrom(type) Then
					Return GenerateDictionary(type, collectionSize, createdObjectReferences)
				End If
			End If

			If type.IsPublic OrElse type.IsNestedPublic Then
				Return GenerateComplexObject(type, createdObjectReferences)
			End If

			Return Nothing
		End Function

		Private Shared Function GenerateTuple(ByVal type As Type, ByVal createdObjectReferences As Dictionary(Of Type, Object)) As Object
			Dim genericArgs() As Type = type.GetGenericArguments()
			Dim parameterValues(genericArgs.Length - 1) As Object
			Dim failedToCreateTuple As Boolean = True
			Dim objectGenerator As New ObjectGenerator()
			For i As Integer = 0 To genericArgs.Length - 1
				parameterValues(i) = objectGenerator.GenerateObject(genericArgs(i), createdObjectReferences)
				failedToCreateTuple = failedToCreateTuple And parameterValues(i) Is Nothing
			Next i
			If failedToCreateTuple Then
				Return Nothing
			End If
			Dim result As Object = Activator.CreateInstance(type, parameterValues)
			Return result
		End Function

		Private Shared Function IsTuple(ByVal genericTypeDefinition As Type) As Boolean
			Return genericTypeDefinition Is GetType(Tuple(Of )) OrElse genericTypeDefinition Is GetType(Tuple(Of ,)) OrElse genericTypeDefinition Is GetType(Tuple(Of ,,)) OrElse genericTypeDefinition Is GetType(Tuple(Of ,,,)) OrElse genericTypeDefinition Is GetType(Tuple(Of ,,,,)) OrElse genericTypeDefinition Is GetType(Tuple(Of ,,,,,)) OrElse genericTypeDefinition Is GetType(Tuple(Of ,,,,,,)) OrElse genericTypeDefinition Is GetType(Tuple(Of ,,,,,,,))
		End Function

		Private Shared Function GenerateKeyValuePair(ByVal keyValuePairType As Type, ByVal createdObjectReferences As Dictionary(Of Type, Object)) As Object
			Dim genericArgs() As Type = keyValuePairType.GetGenericArguments()
			Dim typeK As Type = genericArgs(0)
			Dim typeV As Type = genericArgs(1)
			Dim objectGenerator As New ObjectGenerator()
			Dim keyObject As Object = objectGenerator.GenerateObject(typeK, createdObjectReferences)
			Dim valueObject As Object = objectGenerator.GenerateObject(typeV, createdObjectReferences)
			If keyObject Is Nothing AndAlso valueObject Is Nothing Then
				' Failed to create key and values
				Return Nothing
			End If
			Dim result As Object = Activator.CreateInstance(keyValuePairType, keyObject, valueObject)
			Return result
		End Function

		Private Shared Function GenerateArray(ByVal arrayType As Type, ByVal size As Integer, ByVal createdObjectReferences As Dictionary(Of Type, Object)) As Object
			Dim type As Type = arrayType.GetElementType()
			Dim result As Array = Array.CreateInstance(type, size)
			Dim areAllElementsNull As Boolean = True
			Dim objectGenerator As New ObjectGenerator()
			For i As Integer = 0 To size - 1
				Dim element As Object = objectGenerator.GenerateObject(type, createdObjectReferences)
				result.SetValue(element, i)
				areAllElementsNull = areAllElementsNull And element Is Nothing
			Next i

			If areAllElementsNull Then
				Return Nothing
			End If

			Return result
		End Function

		Private Shared Function GenerateDictionary(ByVal dictionaryType As Type, ByVal size As Integer, ByVal createdObjectReferences As Dictionary(Of Type, Object)) As Object
			Dim typeK As Type = GetType(Object)
			Dim typeV As Type = GetType(Object)
			If dictionaryType.IsGenericType Then
				Dim genericArgs() As Type = dictionaryType.GetGenericArguments()
				typeK = genericArgs(0)
				typeV = genericArgs(1)
			End If

			Dim result As Object = Activator.CreateInstance(dictionaryType)
			Dim addMethod As MethodInfo = If(dictionaryType.GetMethod("Add"), dictionaryType.GetMethod("TryAdd"))
			Dim containsMethod As MethodInfo = If(dictionaryType.GetMethod("Contains"), dictionaryType.GetMethod("ContainsKey"))
			Dim objectGenerator As New ObjectGenerator()
			For i As Integer = 0 To size - 1
				Dim newKey As Object = objectGenerator.GenerateObject(typeK, createdObjectReferences)
				If newKey Is Nothing Then
					' Cannot generate a valid key
					Return Nothing
				End If

				Dim containsKey As Boolean = CBool(containsMethod.Invoke(result, New Object() { newKey }))
				If Not containsKey Then
					Dim newValue As Object = objectGenerator.GenerateObject(typeV, createdObjectReferences)
					addMethod.Invoke(result, New Object() { newKey, newValue })
				End If
			Next i

			Return result
		End Function

		Private Shared Function GenerateEnum(ByVal enumType As Type) As Object
			Dim possibleValues As Array = System.Enum.GetValues(enumType)
			If possibleValues.Length > 0 Then
				Return possibleValues.GetValue(0)
			End If
			Return Nothing
		End Function

		Private Shared Function GenerateQueryable(ByVal queryableType As Type, ByVal size As Integer, ByVal createdObjectReferences As Dictionary(Of Type, Object)) As Object
			Dim isGeneric As Boolean = queryableType.IsGenericType
			Dim list As Object
			If isGeneric Then
				Dim listType As Type = GetType(List(Of )).MakeGenericType(queryableType.GetGenericArguments())
				list = GenerateCollection(listType, size, createdObjectReferences)
			Else
				list = GenerateArray(GetType(Object()), size, createdObjectReferences)
			End If
			If list Is Nothing Then
				Return Nothing
			End If
			If isGeneric Then
				Dim argumentType As Type = GetType(IEnumerable(Of )).MakeGenericType(queryableType.GetGenericArguments())
				Dim asQueryableMethod As MethodInfo = GetType(Queryable).GetMethod("AsQueryable", { argumentType })
				Return asQueryableMethod.Invoke(Nothing, { list })
			End If

			Return Queryable.AsQueryable(CType(list, IEnumerable))
		End Function

		Private Shared Function GenerateCollection(ByVal collectionType As Type, ByVal size As Integer, ByVal createdObjectReferences As Dictionary(Of Type, Object)) As Object
			Dim type As Type = If(collectionType.IsGenericType, collectionType.GetGenericArguments()(0), GetType(Object))
			Dim result As Object = Activator.CreateInstance(collectionType)
			Dim addMethod As MethodInfo = collectionType.GetMethod("Add")
			Dim areAllElementsNull As Boolean = True
			Dim objectGenerator As New ObjectGenerator()
			For i As Integer = 0 To size - 1
				Dim element As Object = objectGenerator.GenerateObject(type, createdObjectReferences)
				addMethod.Invoke(result, New Object() { element })
				areAllElementsNull = areAllElementsNull And element Is Nothing
			Next i

			If areAllElementsNull Then
				Return Nothing
			End If

			Return result
		End Function

		Private Shared Function GenerateNullable(ByVal nullableType As Type, ByVal createdObjectReferences As Dictionary(Of Type, Object)) As Object
			Dim type As Type = nullableType.GetGenericArguments()(0)
			Dim objectGenerator As New ObjectGenerator()
			Return objectGenerator.GenerateObject(type, createdObjectReferences)
		End Function

		Private Shared Function GenerateComplexObject(ByVal type As Type, ByVal createdObjectReferences As Dictionary(Of Type, Object)) As Object
			Dim result As Object = Nothing

			If createdObjectReferences.TryGetValue(type, result) Then
				' The object has been created already, just return it. This will handle the circular reference case.
				Return result
			End If

			If type.IsValueType Then
				result = Activator.CreateInstance(type)
			Else
				Dim defaultCtor As ConstructorInfo = type.GetConstructor(Type.EmptyTypes)
				If defaultCtor Is Nothing Then
					' Cannot instantiate the type because it doesn't have a default constructor
					Return Nothing
				End If

				result = defaultCtor.Invoke(New Object(){})
			End If
			createdObjectReferences.Add(type, result)
			SetPublicProperties(type, result, createdObjectReferences)
			SetPublicFields(type, result, createdObjectReferences)
			Return result
		End Function

		Private Shared Sub SetPublicProperties(ByVal type As Type, ByVal obj As Object, ByVal createdObjectReferences As Dictionary(Of Type, Object))
			Dim properties() As PropertyInfo = type.GetProperties(BindingFlags.Public Or BindingFlags.Instance)
			Dim objectGenerator As New ObjectGenerator()
			For Each [property] As PropertyInfo In properties
				If [property].CanWrite Then
					Dim propertyValue As Object = objectGenerator.GenerateObject([property].PropertyType, createdObjectReferences)
					[property].SetValue(obj, propertyValue, Nothing)
				End If
			Next [property]
		End Sub

		Private Shared Sub SetPublicFields(ByVal type As Type, ByVal obj As Object, ByVal createdObjectReferences As Dictionary(Of Type, Object))
			Dim fields() As FieldInfo = type.GetFields(BindingFlags.Public Or BindingFlags.Instance)
			Dim objectGenerator As New ObjectGenerator()
			For Each field As FieldInfo In fields
				Dim fieldValue As Object = objectGenerator.GenerateObject(field.FieldType, createdObjectReferences)
				field.SetValue(obj, fieldValue)
			Next field
		End Sub

		Private Class SimpleTypeObjectGenerator
			Private _index As Long = 0
			Private Shared ReadOnly DefaultGenerators As Dictionary(Of Type, Func(Of Long, Object)) = InitializeGenerators()

			<SuppressMessage("Microsoft.Maintainability", "CA1502:AvoidExcessiveComplexity", Justification := "These are simple type factories and cannot be split up.")>
			Private Shared Function InitializeGenerators() As Dictionary(Of Type, Func(Of Long, Object))
                Return New Dictionary(Of Type, Func(Of Long, Object)) From {
                {GetType(Boolean), Function(index) True},
                {GetType(Byte), Function(index) CType(64, Byte)},
                {GetType(Char), Function(index) ChrW(65)},
                {GetType(Date), Function(index) Date.Now},
                {GetType(DateTimeOffset), Function(index) New DateTimeOffset(Date.Now)},
                {GetType(DBNull), Function(index) DBNull.Value},
                {GetType(Decimal), Function(index) CType(index, Decimal)},
                {GetType(Double), Function(index) CType(index + 0.1, Double)},
                {GetType(Guid), Function(index) Guid.NewGuid()},
                {GetType(Int16), Function(index) CShort(Fix(index Mod Int16.MaxValue))},
                {GetType(Int32), Function(index) CInt(Fix(index Mod Int32.MaxValue))},
                {GetType(Int64), Function(index) CLng(Fix(index))},
                {GetType(Object), Function(index) New Object()},
                {GetType(SByte), Function(index) CType(64, SByte)},
                {GetType(Single), Function(index) CType(index + 0.1, Single)},
                {GetType(String), Function(index) String.Format(CultureInfo.CurrentCulture, "sample string {0}", index)},
                {GetType(TimeSpan), Function(index) TimeSpan.FromTicks(1234567)},
                {GetType(UInt16), Function(index) CUShort(index Mod UInt16.MaxValue)},
                {GetType(UInt32), Function(index) CUInt(index Mod UInt32.MaxValue)},
                {GetType(UInt64), Function(index) CULng(index)},
                {GetType(Uri), Function(index) New Uri(String.Format(CultureInfo.CurrentCulture, "http://webapihelppage{0}.com", index))}
                }

            End Function

            Public Shared Function CanGenerateObject(ByVal type As Type) As Boolean
				Return DefaultGenerators.ContainsKey(type)
			End Function

			Public Function GenerateObject(ByVal type As Type) As Object
				_index += 1
				Return DefaultGenerators(type)(_index)
			End Function
		End Class
	End Class
End Namespace